/************************************************************
** $Id: database.c,v 0.6 2003/11/24 18:38:35 bcx Exp $
**
** Copyright Bryan Costales and Perforce Software, Inc. 2003
**
** This code is "open source" as defined by version 1.9 of
** the Open Source Definition from:
**
**	http://www.opensource.org/docs/definition.php.
**
** Entry points defined in this file:
**
** open_database():
** 	Open a database, creating it if asked to.
** close_database():
** 	Close a database.
** put_ip_database():
** 	Add a date to the primary database of IP numbers.
**	Create a new entry of the IP number is not found.
** get_ip_database():
** 	Look up an ip number in the primary database
** del_ip_database():
** 	Remove and IP record and all its events
**	or just one date's record.
** put_event_database():
** 	Add an ip/date item to the secondary database.
** get_event_database():
** 	Read an ip/date item from the secondary database.
** dump_database():
** 	Print the database contents to a file pointer.
** event_to_str():
** 	Format the event for printing or logging.
*************************************************************/

# define EXTERN extern
# include "slow.h"

DB *
open_database(char *fname, int force)
{
	static DB *dp;
	u_int32_t flags = DB_THREAD;
	int ret;

	ret = db_create(&dp, (DB_ENV *)NULL, 0);
	if (ret != 0)
	{
		errno = ret;
		return NULL;
	}
	if (force == TRUE)
		flags |= DB_CREATE;
	ret = dp->open(dp, NULL, fname, NULL, DB_HASH, flags, 0644);
	if (ret != 0)
	{
		errno = ret;
		return NULL;
	}
	return dp;
}

void
close_database(DB *dp)
{
	(void)dp->close(dp, 0);
}

/*
** Append a new date to the list of dates for a given IP address.
** The dates will be in ascending order, with the oldest at the
** start. This scheme will only break if the date is manually
** reset on the machine (and even worse if the date is set backwards).
**
** If the IP key doesn't exist yet, it is begun with a single date
** record. The IP/date combination must be added to the secondary
** database for this to work. An unreferenced IP/date key in that
** database will produce a warning.
*/
int
put_ip_database(DB *dp, PRIMARY_KEY ip, PRIMARY_DATUM date, bool drop, int log)
{
	DBT key, data;
	int ret;
	PRIMARY_DATUM *new, *d, *n;
	int i, got;
	extern int errno;

	if (dp == NULL)
		return EFAULT;

	(void) memset(&data, 0, sizeof data);
	(void) memset(&key, 0, sizeof key);

	key.data = &ip;
	key.size = sizeof(PRIMARY_KEY);
	data.data = NULL;
	data.size = 0;
	data.ulen = 0;
	data.flags = DB_DBT_MALLOC;

	ret = dp->get(dp, NULL, &key, &data, 0);
	if (ret != 0)
	{
		if (ret != DB_NOTFOUND)
		{
			return ret;
		}
		if (drop == TRUE)
			return DB_NOTFOUND;
		data.size = 0;
		data.data = try_alloc(__FILE__, __LINE__,
			sizeof(PRIMARY_DATUM), log);
		if (data.data == NULL)
			return DB_NOTFOUND;
	}
	else
	{
		if (drop == FALSE)
		{
			/*
			 * Is the date already in the database for this
			 * IP number. If so, just update the record.
			 */
			got = FALSE;
			d = (PRIMARY_DATUM *)data.data;
			for (i = 0; i < data.size; i += sizeof(PRIMARY_DATUM))
			{
				if (*d == date)
				{
					got = TRUE;
					break;
				}
				++d;
			}
			if (got == TRUE)
				return 0;
			data.data = try_realloc(__FILE__, __LINE__, data.data,
				data.size + sizeof(PRIMARY_DATUM), log);
			if (data.data == NULL)
				return DB_NOTFOUND;
		}
		else
		{
			new = (PRIMARY_DATUM *)try_alloc(__FILE__, __LINE__, data.size, log);
			n = new;
			if (n == NULL)
				return DB_NOTFOUND;
			d = (PRIMARY_DATUM *)data.data;
			got = FALSE;
			for (i = 0; i < data.size; i += sizeof(PRIMARY_DATUM))
			{
				if (*d == date)
				{
					++d;
					got = TRUE;
					continue;
				}
				*n = *d;
				++n;
				++d;
			}
			if (got == FALSE)
			{
				(void) free(data.data);
				(void) free(new);
				return DB_NOTFOUND;
			}
			(void) free(data.data);
			data.data = new;
			data.size = data.size - sizeof(PRIMARY_DATUM);
		}
	}
	if (drop == FALSE)
	{
		(void)memcpy(((char *)(data.data)) + data.size, &date, sizeof(PRIMARY_DATUM));
		data.size += sizeof(PRIMARY_DATUM);
	}

	if (data.size > 0)
		ret = dp->put(dp, NULL, &key, &data, 0);
	else
		ret = dp->del(dp, NULL, &key, 0);

	if (ret != 0)
	{
		(void)free(data.data);
		return ret;
	}
	(void)dp->sync(dp, 0);

	(void)free(data.data);
	return 0;
}

/*
** Look up an IP number in the primary database and return
** its list of dates. The list is in allocated memory, so 
** it can be freed.
** 
** If the IP number is not found, return NULL and set ndates to 0.
** On error return NULL, and sets ndates to non-zero.
** No data returns a valid address but set ndates to 0.
** Otherwise return address of list of dates and set ndates to how many dates.
*/
PRIMARY_DATUM *
get_ip_database(DB *dp, PRIMARY_KEY ip, int *ndates, int log)
{
	DBT key, data;
	int ret;
	extern int errno;

	if (dp == NULL)
	{
		errno = *ndates = EFAULT;
		return NULL;
	}

	(void) memset(&key, 0, sizeof key);
	(void) memset(&data, 0, sizeof data);

	key.data = &ip;
	key.size = sizeof(PRIMARY_KEY);
	data.data = NULL;
	data.size = 0;
	data.ulen = 0;
	data.flags = DB_DBT_MALLOC;

	ret = dp->get(dp, NULL, &key, &data, 0);
	if (ret != 0)
	{
		if (ret == DB_NOTFOUND)
		{
			*ndates = 0;
			return NULL;
		}
		errno = *ndates = ret;
		data.data = try_alloc(__FILE__, __LINE__,
			sizeof(PRIMARY_DATUM), log);
		return data.data;
	}
	*ndates = data.size / sizeof(PRIMARY_DATUM);
	return data.data;
}

/*
** Delete a record for an IP number or the entire
** IP number and all its records.
** 
** If the IP number is not found, return FALSE.
** On error return FALSE.
** If the date is zero (no date specified) remove everything.
*/
int
del_ip_database(DB *dp1, DB *dp2, PRIMARY_KEY ip, time_t date, int log)
{
	SECONDARY_KEY k;
	PRIMARY_DATUM *dates, *d;
	int ndates, i, ret;
	extern int errno;

	if (dp1 == NULL || dp2 == NULL)
		return (errno = EINVAL);

	errno = 0;
	ndates = 0;
	if ((dates = get_ip_database(dp1, ip, &ndates, log)) == NULL)
	{
		if (ndates == 0)
		{
			errno = EINVAL;
			return DB_NOTFOUND;
		}
		return ndates;
	}
	if (date != 0)
	{
		for(d = dates, i = 0; i < ndates; ++i, ++d)
		{
			if (date == *d)
				break;
		}
		if (i == ndates)
			return DB_NOTFOUND;
		errno = 0;
		ret = put_ip_database(dp1, ip, date, TRUE, log);
		if (ret != 0)
			return ret;
		k.ip = ip;
		k.date = date;
		errno = 0;
		return del_event_database(dp2, &k);
	}
	for(d = dates, i = 0; i < ndates; ++i, ++d)
	{
		errno = 0;
		ret = put_ip_database(dp1, ip, *d, TRUE, log);
		if (ret != 0)
			return ret;
		k.ip = ip;
		k.date = *d;
		errno = 0;
		ret = del_event_database(dp2, &k);
		if (ret != 0)
			return ret;
	}

	return 0;
}

int
put_event_database(DB *dp, SECONDARY_KEY *k, SECONDARY_DATUM *event)
{
	DBT key, data;
	int ret;
	extern int errno;

	if (dp == NULL)
		return EFAULT;

	(void) memset(&data, 0, sizeof data);
	(void) memset(&key, 0, sizeof key);

	key.data = k;
	key.size = sizeof(SECONDARY_KEY);
	data.data = event;
	data.size = sizeof(SECONDARY_DATUM);

	/*
	** If IP,date already exists as a key, this
	** will silently overwrite the old entry.
	*/
	ret = dp->put(dp, NULL, &key, &data, 0);
	if (ret != 0)
	{
		return ret;
	}
	(void)dp->sync(dp, 0);

	return 0;
}

int
del_event_database(DB *dp, SECONDARY_KEY *k)
{
	DBT key, data;
	int ret;
	extern int errno;

	if (dp == NULL)
		return EFAULT;

	(void) memset(&data, 0, sizeof data);
	(void) memset(&key, 0, sizeof key);

	key.data = k;
	key.size = sizeof(SECONDARY_KEY);

	/*
	** If IP,date already exists as a key, this
	** will silently overwrite the old entry.
	*/
	ret = dp->del(dp, NULL, &key, 0);
	if (ret != 0)
	{
		return ret;
	}
	(void)dp->sync(dp, 0);

	return 0;
}

SECONDARY_DATUM *
get_event_database(DB *dp, SECONDARY_KEY *k, bool verbose)
{
	DBT key, data;
	int ret;
	extern int errno;

	if (dp == NULL)
		return NULL;

	(void) memset(&key, 0, sizeof key);
	(void) memset(&data, 0, sizeof data);

	key.data = k;
	key.size = sizeof(SECONDARY_KEY);
	data.data = NULL;
	data.size = 0;
	data.ulen = 0;
	data.flags = DB_DBT_MALLOC;

	ret = dp->get(dp, NULL, &key, &data, 0);
	if (ret != 0)
	{
		if (verbose)
		{
			dp->err(dp, ret, "%s", SecondaryDatabaseLoc);
		}
		errno = ret;
		return NULL;
	}

	return (SECONDARY_DATUM *)(data.data);
}

/*
** Walk the databases, producing a report of its contents.
** The fp pointer is where to print the report, and is usually
** the standard output.
**
** This routine is intended for use from the command-line, and is
** not thread safe.
*/
int
dump_database(DB *ipdp, DB *recdp, FILE *fp)
{
	DBT key, data;
	int i, ret, records;
	DBC *curs;
	SECONDARY_DATUM *event;
	SECONDARY_KEY k;
	struct in_addr in;
	time_t *tp;
	char ipbuf[MAXHOSTNAMELEN];
	extern int errno;

	if (ipdp == NULL)
		return 0;

	ret = ipdp->cursor(ipdp, NULL, &curs, 0);
	if (ret != 0)
	{
		return ret;
	}

	(void) memset(&data, 0, sizeof data);
	(void) memset(&key, 0, sizeof key);
	key.flags = DB_DBT_REALLOC;
	data.flags = DB_DBT_REALLOC;
	for (;;)
	{
		ret = curs->c_get(curs, &key, &data, DB_NEXT);
		if (ret == DB_NOTFOUND)
				break;
		if (ret != 0)
		{
			(void) curs->c_close(curs);
			return ret;
		}
		if (key.data == NULL)
			continue;
		records = data.size / sizeof(PRIMARY_DATUM);
		(void) memcpy(&in.s_addr, ((char *)(key.data)), sizeof(in.s_addr));
# ifndef WORDS_BIGENDIAN
		in.s_addr = revbytes(in.s_addr);
# endif
		(void) sprintf(ipbuf, "%s", inet_ntoa(in));

		tp = (time_t *)data.data;
		for (i = 0; i < records; i++, tp++)
		{
			k.ip = *((PRIMARY_KEY *)(key.data));
			k.date = *tp;
			event = get_event_database(recdp, &k, FALSE);
			fprintf(fp, "%s\n", event_to_str(ipbuf, k.date, event, FALSE));
			(void) free((char *)event);
		}
	}
	(void) curs->c_close(curs);
	return 0;
}

char *
event_to_str(char *ipbuf, time_t date, SECONDARY_DATUM *event, int log)
{
	char buf[BUFSIZ];
	struct tm *t;

	if (ipbuf == NULL)
		return NULL;

	t = localtime(&date);
	(void) sprintf(buf, "%s: %02d/%02d/%d.%02d:%02d:%02d ",
		ipbuf,
		t->tm_mon + 1, t->tm_mday, t->tm_year + 1900,
		t->tm_hour, t->tm_min, t->tm_sec);
	if (event == NULL)
	{
		(void) sprintf(buf+strlen(buf), "%s",
			errno == DB_NOTFOUND ? "Item not in database (need to garbage collect)" :
			"Lookup error while processing item");
		return try_strdup(__FILE__, __LINE__, buf, log);
	}
	(void) sprintf(buf+strlen(buf), "envrcpts=%d ", event->envrcpts); 
	(void) sprintf(buf+strlen(buf), "hdrrcpts=%d ", event->headrcpts);
	(void) sprintf(buf+strlen(buf), "honeyrcpts=%d ", event->honeyrcpts);
	(void) sprintf(buf+strlen(buf), "bad=%d ", event->badrcpts);
	if (event->eventmap == 0)
	{
		(void) strcat(buf, "bits=0");
	}
	else
	{
		(void) strcat(buf, "bits=");
		if (isbitset(event->eventmap, BIT_TOO_MANY_HDR_RCPTS))
			(void) strcat(buf, "[HED]");
		if (isbitset(event->eventmap, BIT_TOO_MANY_ENV_RCPTS))
			(void) strcat(buf, "[ENV]");
		if (isbitset(event->eventmap, BIT_GOT_HONEYPOT))
		(void) strcat(buf, "[HON]");
		if (isbitset(event->eventmap, BIT_BAD_MSG_ID))
			(void) strcat(buf, "[MID]");
		if (isbitset(event->eventmap, BIT_BAD_FROM_HEAD))
			(void) strcat(buf, "[FRM]");
		if (isbitset(event->eventmap, BIT_BAD_HOST))
		(void) strcat(buf, "[BHO]");
		if (isbitset(event->eventmap, BIT_RBL_BAD))
			(void) strcat(buf, "[RBL]");
		if (isbitset(event->eventmap, BIT_MILTER_ABORTED))
			(void) strcat(buf, "[ABO]");
		if (isbitset(event->eventmap, BIT_ADVANCEWRITE))
		(void) strcat(buf, "[PIP]");
		if (isbitset(event->eventmap, BIT_FORGED))
			(void) strcat(buf, "[FOR]");
		if (isbitset(event->eventmap, BIT_FROMMX))
			(void) strcat(buf, "[MMX]");
		if (isbitset(event->eventmap, BIT_NOTINET))
			(void) strcat(buf, "[NET]");
	}
	(void) sprintf(buf+strlen(buf), " msgid=%s", event->msgid);
	
	return try_strdup(__FILE__, __LINE__, buf, log);
}
